/*
 * ImageProjection.java  1.0  98/06/03  Carl Burke
 *
 * Base class for generating images using user defined procedures;
 * intended application is generation of synthetic terrain and overhead
 * maps.  This is a simple shell.
 *
 * Copyright (c) 1998 Carl Burke.  All rights reserved.
 */

import java.awt.*;
import java.awt.image.*;
import java.util.*;

public abstract class ImageProjection implements ImageProducer, Runnable
{
    boolean imageDone;
    int Width;
    int Height;
    int[] pixels;
    ColorModel model;
    Hashtable properties;

    SolidNoiseGenerator SNG;

    double longi,lat;
    double vgrid, hgrid;

    double scale;

    ///** SUPPORT ROUTINES ***

    int min(int x,int y)
    { return(x<y ? x : y); }

    int max(int x,int y)
    { return(x<y ? y : x); }

    double log2(double x)
    {
	return(Math.log(x)/Math.log(2.0));
    }

    public static final double DEG2RAD = 0.017453292; // pi/180

    public ImageProjection(int w, int h, SolidNoiseGenerator sng)
    {
	consumers = new Vector();
	imageDone = false;
	Width = w;
	Height = h;
	pixels = new int[Width*Height];
	model = ColorModel.getRGBdefault();
	properties = new Hashtable();
	longi = 0.0;
	lat = 0.0;
	vgrid = 0.0;
	hgrid = 0.0;
	SNG = sng;
	setMagnification(1.0);
	setCenterOfProjection(0.0, 0.0);
    }

    public void forcePixelRecalc()
    {
	imageDone = false;
    }

    public void setMagnification(double m)
    {
	scale = m;
    }
    public double getMagnification() {return scale;}

    double cla, sla, clo, slo;
    public void setCenterOfProjection(double lg, double lt)
    {
	longi = lg;
	lat = lt;
	sla = Math.sin(lat); cla = Math.cos(lat);
	slo = Math.sin(longi); clo = Math.cos(longi);
    }
    public double getLongitudeCOP() {return longi;}
    public double getLatitudeCOP() {return lat;}

    public void setGridSpacing(double h, double v)
    {
	hgrid = h;
	vgrid = v;
    }
    public double getHGridSpacing() {return hgrid;}
    public double getVGridSpacing() {return vgrid;}

    private Vector consumers = null;

    /**
     * Adds an ImageConsumer to the list of consumers interested in
     * data for this image.
     * @see ImageConsumer
     */
    public synchronized void addConsumer(ImageConsumer ic)
    {
	if (isConsumer(ic) == false) consumers.addElement(ic);
    }

    /**
     * Remove an ImageConsumer from the list of consumers interested in
     * data for this image.
     * @see ImageConsumer
     */
    public synchronized void removeConsumer(ImageConsumer ic)
    {
	if (isConsumer(ic) == true) consumers.removeElement(ic);
    }

    /**
     * Determine if an ImageConsumer is on the list of consumers currently
     * interested in data for this image.
     * @return true if the ImageConsumer is on the list; false otherwise
     * @see ImageConsumer
     */
    public boolean isConsumer(ImageConsumer ic)
    {
	return consumers.indexOf(ic) > -1;
    }

    /**
     * Requests that a given ImageConsumer have the image data delivered
     * one more time in top-down, left-right order.
     * @see ImageConsumer
     *
     * We don't observe this request, since data is always sent TDLR.
     */
    public void requestTopDownLeftRightResend(ImageConsumer ic)
    {
	;
    }
 
    /**
     * Image production starts here.  Image production occurs in a
     * separate thread.  The application itself is free to go on
     * with life.
     */

    private Thread projectThread = null;

    /**
     * Adds an ImageConsumer to the list of consumers interested in
     * data for this image, and immediately start delivery of the
     * image data through the ImageConsumer interface.
     * @see ImageConsumer
     */
    public void startProduction(ImageConsumer ic)
    {
	addConsumer(ic);
    
	if (imageDone)
	{
	    sendPixels(ic);	// send finished image to this consumer
	}
	else if (projectThread == null)
	{
	    projectThread = new Thread(this);
	    projectThread.start();
	}
	// image isn't done, but thread is running; since we clone the
	// consumers vector in 'run()', this consumer may lose out...
	// need to think about this a little more.
    }

    /**
     * This routine is called when the data have already been generated;
     * we can send pixels to the ImageConsumer immediately.
     */
    private void sendPixels(ImageConsumer ic)
    {
        ic.setDimensions(Width, Height);
        ic.setColorModel(model);
	ic.setHints(ImageConsumer.TOPDOWNLEFTRIGHT |
		    ImageConsumer.COMPLETESCANLINES |
		    ImageConsumer.SINGLEPASS |
		    ImageConsumer.SINGLEFRAME);
        ic.setPixels(0, 0, Width, Height, model, pixels, 0, Width);
        ic.imageComplete(ImageConsumer.STATICIMAGEDONE);
    }

    /**
     * This is the image production controller.
     */

    public void run()
    {
	Vector v = (Vector)consumers.clone();

	Enumeration e;

	// Tell each image consumer the dimensions of the image
	//   and the color model we will use.  Also inform the
	//   consumers that we expect to send a single frame, TDLR,
	//   complete in one go.

	for (e = v.elements(); e.hasMoreElements(); )
	{
	    ImageConsumer ic = (ImageConsumer)e.nextElement();
	    
	    ic.setDimensions(Width, Height);
	    ic.setColorModel(model);
	    ic.setHints(ImageConsumer.TOPDOWNLEFTRIGHT |
		        ImageConsumer.COMPLETESCANLINES |
		        ImageConsumer.SINGLEPASS |
		        ImageConsumer.SINGLEFRAME);
	}

	// Do the generation according to the implemented projection
	renderTerrain();
	if (hgrid > 0.0) renderLatitudes();
	if (vgrid > 0.0) renderLongitudes();
	imageDone = true;

	// Supply pixels to each image consumer, now that the image is finished.

	for (e = v.elements(); e.hasMoreElements(); )
	{
	    ImageConsumer ic = (ImageConsumer)e.nextElement();
	    
	    ic.setPixels(0, 0, Width, Height, model, pixels, 0, Width);
	    ic.imageComplete(ImageConsumer.STATICIMAGEDONE);
	}

	projectThread = null;
    }

    /**
     * Every subclass must implement these functions to create the image.
     */
    public abstract void renderTerrain();
    public abstract void renderLatitudes();
    public abstract void renderLongitudes();
}
